@@ -72,6 +72,34 @@ extern "C" {
 #define BTNAME_MATCH_RULE "type='signal',interface='" BTNAME_SIGNAL_IF \
                           "',member='" BTNAME_SIG_CHANGED "'"
 
+static GSList *dialog_parents = NULL;
+
+GtkWindow *
+get_dialog_parent ()
+{
+  if (dialog_parents)
+    return GTK_WINDOW (dialog_parents->data);
+  else
+    return NULL;
+}
+
+void
+push_dialog_parent (GtkWidget *w)
+{
+  dialog_parents = g_slist_prepend (dialog_parents, w);
+}
+
+void
+pop_dialog_parent ()
+{
+  if (dialog_parents)
+    {
+      GSList *old = dialog_parents;
+      dialog_parents = dialog_parents->next;
+      g_slist_free_1 (old);
+    }
+}
+
 struct ayn_closure {
   package_info *pi;
   detail_kind kind;
@@ -88,7 +116,7 @@ yes_no_response (GtkDialog *dialog, gint response, gpointer clos)
   if (response == 1)
     {
       if (c->pi)
-	show_package_details (GTK_WINDOW (dialog), c->pi, c->kind, false);
+	show_package_details (c->pi, c->kind, false);
       else if (c->details)
 	c->details (c->data);
       return;
@@ -100,6 +128,7 @@ yes_no_response (GtkDialog *dialog, gint response, gpointer clos)
     c->pi->unref ();
   delete c;
 
+  pop_dialog_parent ();
   gtk_widget_destroy (GTK_WIDGET (dialog));
   cont (response == GTK_RESPONSE_OK, data);
 }
@@ -116,7 +145,8 @@ ask_yes_no (const gchar *question,
   c->details = NULL;
   c->data = data;
 
-  dialog = hildon_note_new_confirmation (get_main_window (), question);
+  dialog = hildon_note_new_confirmation (get_dialog_parent (), question);
+  push_dialog_parent (dialog);
 
   g_signal_connect (dialog, "response",
 		    G_CALLBACK (yes_no_response), c);
@@ -137,11 +167,12 @@ ask_custom (const gchar *question,
   c->data = data;
 
   dialog = hildon_note_new_confirmation_add_buttons 
-    (get_main_window (),
+    (get_dialog_parent (),
      question,
      ok_label, GTK_RESPONSE_OK,
      cancel_label, GTK_RESPONSE_CANCEL,
      NULL);
+  push_dialog_parent (dialog);
 
   g_signal_connect (dialog, "response",
 		    G_CALLBACK (yes_no_response), c);
@@ -166,12 +197,13 @@ ask_yes_no_with_details (const gchar *title,
 
   dialog = gtk_dialog_new_with_buttons
     (title,
-     get_main_window (),
+     get_dialog_parent (),
      GTK_DIALOG_MODAL,
      _("ai_bd_confirm_ok"),      GTK_RESPONSE_OK,
      _("ai_bd_confirm_details"), 1,
      _("ai_bd_confirm_cancel"),  GTK_RESPONSE_CANCEL,
      NULL);
+  push_dialog_parent (dialog);
 
   gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
   gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox),
@@ -198,12 +230,13 @@ ask_yes_no_with_arbitrary_details (const gchar *title,
 
   dialog = gtk_dialog_new_with_buttons
     (title,
-     get_main_window (),
+     get_dialog_parent (),
      GTK_DIALOG_MODAL,
      _("ai_bd_confirm_ok"),      GTK_RESPONSE_OK,
      _("ai_bd_confirm_details"), 1,
      _("ai_bd_confirm_cancel"),  GTK_RESPONSE_CANCEL,
      NULL);
+  push_dialog_parent (dialog);
 
   gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
   GtkWidget *label = gtk_label_new (question);
@@ -221,6 +254,7 @@ static bool currently_annoying_user = false;
 static void
 annoy_user_response (GtkDialog *dialog, gint response, gpointer data)
 {
+  pop_dialog_parent ();
   gtk_widget_destroy (GTK_WIDGET (dialog));
   currently_annoying_user = false;
 }
@@ -233,7 +267,8 @@ annoy_user (const gchar *text)
 
   GtkWidget *dialog;
 
-  dialog = hildon_note_new_information (get_main_window (), text);
+  dialog = hildon_note_new_information (get_dialog_parent (), text);
+  push_dialog_parent (dialog);
   g_signal_connect (dialog, "response", 
 		    G_CALLBACK (annoy_user_response), NULL);
   gtk_widget_show_all (dialog);
@@ -253,10 +288,11 @@ annoy_user_with_details_response (GtkDialog *dialog, gint response,
 
   if (response == 1)
     {
-      show_package_details (GTK_WINDOW (dialog), c->pi, c->kind, true);
+      show_package_details (c->pi, c->kind, true);
     }
   else
     {
+      pop_dialog_parent ();
       gtk_widget_destroy (GTK_WIDGET (dialog));
       currently_annoying_user = false;
       c->pi->unref ();
@@ -274,7 +310,8 @@ annoy_user_with_details (const gchar *text,
   GtkWidget *dialog;
   auwd_closure *c = new auwd_closure;
 
-  dialog = hildon_note_new_information (get_main_window (), text);
+  dialog = hildon_note_new_information (get_dialog_parent (), text);
+  push_dialog_parent (dialog);
 
   {
     // XXX - the buttons should be "Details" "Close", so we remove the
@@ -310,6 +347,7 @@ static void
 annoy_user_with_log_response (GtkDialog *dialog, gint response,
 			      gpointer data)
 {
+  pop_dialog_parent ();
   gtk_widget_destroy (GTK_WIDGET (dialog));
   currently_annoying_user = false;
 
@@ -325,7 +363,8 @@ annoy_user_with_log (const gchar *text)
 
   GtkWidget *dialog;
 
-  dialog = hildon_note_new_information (get_main_window (), text);
+  dialog = hildon_note_new_information (get_dialog_parent (), text);
+  push_dialog_parent (dialog);
 #if 0
   gtk_dialog_add_button (GTK_DIALOG (dialog), "Log", 1);
 #endif
@@ -399,11 +438,12 @@ scare_user_with_legalese (bool sure,
 
   dialog = gtk_dialog_new_with_buttons
     (_("ai_ti_notice"),
-     get_main_window (),
+     get_dialog_parent (),
      GTK_DIALOG_MODAL,
      _("ai_bd_notice_ok"),      GTK_RESPONSE_OK,
      _("ai_bd_notice_cancel"),  GTK_RESPONSE_CANCEL,
      NULL);
+  push_dialog_parent (dialog);
 
   gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
   const char *text = (sure
@@ -497,6 +537,7 @@ create_progress (const gchar *title, bool with_cancel)
 {
   if (progress_dialog)
     {
+      pop_dialog_parent ();
       gtk_widget_destroy (progress_dialog);
       progress_dialog = NULL;
       progress_bar = NULL;
@@ -516,9 +557,10 @@ create_progress (const gchar *title, bool with_cancel)
   gchar *longer_title = g_strconcat (title, "XXX", NULL);
   progress_bar = GTK_PROGRESS_BAR (gtk_progress_bar_new ());
   progress_dialog =
-    hildon_note_new_cancel_with_progress_bar (get_main_window (),
+    hildon_note_new_cancel_with_progress_bar (get_dialog_parent (),
 					      longer_title,
 					      progress_bar);
+  push_dialog_parent (progress_dialog);
   g_free (longer_title);
   g_object_set (progress_dialog, "description", title, NULL);
 
@@ -531,8 +573,11 @@ create_progress (const gchar *title, bool with_cancel)
       g_list_free (kids);
     }
   else
-    g_signal_connect (progress_dialog, "response",
-		      G_CALLBACK (cancel_response), NULL);
+    {
+      respond_on_escape (GTK_DIALOG (progress_dialog), GTK_RESPONSE_CANCEL);
+      g_signal_connect (progress_dialog, "response",
+			G_CALLBACK (cancel_response), NULL);
+    }
 
   gtk_widget_show (progress_dialog);
 }
@@ -644,6 +689,7 @@ hide_progress ()
   if (progress_dialog)
     {
       stop_pulsing ();
+      pop_dialog_parent ();
       gtk_widget_destroy (GTK_WIDGET(progress_bar));
       gtk_widget_destroy (progress_dialog);
       progress_dialog = NULL;
@@ -1423,6 +1469,7 @@ fcd_response (GtkDialog *dialog, gint response, gpointer clos)
 
   char *uri = gtk_file_chooser_get_uri (GTK_FILE_CHOOSER (dialog));
 
+  pop_dialog_parent ();
   gtk_widget_destroy (GTK_WIDGET (dialog));
 
   if (response == GTK_RESPONSE_OK)
@@ -1443,13 +1490,14 @@ show_deb_file_chooser (void (*cont) (char *uri, void *data),
   GtkFileFilter *filter;
 
   fcd = hildon_file_chooser_dialog_new_with_properties
-    (get_main_window (),
+    (get_dialog_parent (),
      "action",            GTK_FILE_CHOOSER_ACTION_OPEN,
      "title",             _("ai_ti_select_package"),
      "empty_text",        _("ai_ia_select_package_no_packages"),
      "open_button_text",  _("ai_bd_select_package"),
      NULL);
   gtk_window_set_modal (GTK_WINDOW (fcd), TRUE);
+  push_dialog_parent (fcd);
 
   filter = gtk_file_filter_new ();
   gtk_file_filter_add_mime_type (filter, "application/x-deb");
@@ -1478,10 +1526,11 @@ show_file_chooser_for_save (const char *title,
   GtkWidget *fcd;
 
   fcd = hildon_file_chooser_dialog_new_with_properties
-    (parent,
+    (get_dialog_parent (),
      "action",            GTK_FILE_CHOOSER_ACTION_SAVE,
      "title",             title,
      NULL);
+  push_dialog_parent (fcd);
   gtk_window_set_modal (GTK_WINDOW (fcd), TRUE);
 
   gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (fcd), default_filename);
